package net.es.oscars.policy;

import java.rmi.RemoteException;
import java.util.ArrayList;

import org.apache.log4j.Logger;
import org.hibernate.Session;

import net.es.oscars.database.HibernateUtil;
import net.es.oscars.database.Initializer;
import net.es.oscars.policy.rmi.PolicyRmiServer;

/**
 * Core of the policy server that maintains singletons and global methods
 * 
 * @author Andrew Lake (alake@internet2.edu)
 */
public class PolicyCore {
    private static PolicyCore instance = null;
    private Logger log = null;
    public boolean initialized = false;
    private PolicyRmiServer policyRmiServer = null;
    private String policyDbName = "oscars_policy";
    
    /**
     * Constructor - private because this is a Singleton
     */
    private PolicyCore() {
        this.log = Logger.getLogger(this.getClass());
    }

    /**
     * Initializer
     * @return the OSCARSCore singleton instance
     */
    public static PolicyCore init() {
        if (PolicyCore.instance == null) {
            PolicyCore.instance = new PolicyCore();
        }
        PolicyCore instance = PolicyCore.instance;
        instance.initAll();
        return instance;
    }

    private void initAll() {
        this.log.debug("initAll.start");
        this.initPolicyDatabase();
        this.initRMIServer();
        this.log.debug("initAll.end");
    }

    /**
     * Shuts down all modules
     */
    public void shutdown() {
        this.log.info("shutdown.start");
        try {
            this.policyRmiServer.shutdown();
        } catch (Exception e) {
            this.log.error("Caught Exception on policyRmiServer.shutdown", e);
        }
        //TODO: Shut down more stuff
        this.log.info("shutdown.end");
    }

    /**
     * @return the PolicyCore singleton instance
     */
    public static PolicyCore getInstance() {
        if (PolicyCore.instance == null) {
            PolicyCore.instance = new PolicyCore();
        }
        return PolicyCore.instance;
    }

    /**
     * Initializes the RMIServer module
     */
    public void initRMIServer() {
        this.log.info("initRMIServer.start");
        try {
            this.policyRmiServer = new PolicyRmiServer();
            this.policyRmiServer.init();
        } catch (RemoteException ex) {
            this.log.error("Error initializing RMI server", ex);
            this.policyRmiServer.shutdown();
            this.policyRmiServer = null;
            System.out.println("Failure to initialize Policy RMI server, exiting. Exception was " + ex.getMessage());
            this.log.error("PolicyCore exiting");
            System.exit(1);
        }
        this.log.info("initRMIServer.end");
    }
    
    /**
     * Initializes the DB module
     */
    public void initPolicyDatabase() {
        this.log.debug("initPolicyDatabase.start");
        ArrayList<String> dbnames = new ArrayList<String>();
        dbnames.add(this.policyDbName);
        Initializer dbInitializer = new Initializer();
        dbInitializer.initDatabase(dbnames);
        this.log.debug("initPolicyDatabase.end");
    }
    
    /**
     * @return Grabs the current Policy DB session for the current thread
     */
    public Session getPolicySession() {
        Session policy = HibernateUtil.getSessionFactory(this.policyDbName).getCurrentSession();
        if (policy == null || !policy.isOpen()) {
            this.log.debug("opening policy session");
            policy = HibernateUtil.getSessionFactory(this.policyDbName).openSession();
            policy = HibernateUtil.getSessionFactory(this.policyDbName).getCurrentSession();
        }
        if (policy == null || !policy.isOpen()) {
            this.log.error("Policy session is still closed!");
        }
        return policy;
    }

    /**
     * @return the policyDbName
     */
    public String getPolicyDbName() {
        return this.policyDbName;
    }
}
