package net.es.oscars.policy;

import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Properties;

import net.es.oscars.PropHandler;
import net.es.oscars.aaa.Attribute;
import net.es.oscars.aaa.User;
import net.es.oscars.policy.client.CheckPolicyRequest;
import net.es.oscars.policy.client.CheckPolicyResponse;
import net.es.oscars.policy.client.Requester;
import net.es.oscars.policy.database.*;
import net.es.oscars.rmi.aaa.AaaRmiClient;
import net.es.oscars.rmi.model.ModelObject;
import net.es.oscars.rmi.model.ModelOperation;

import org.apache.log4j.Logger;
import org.opensaml.saml2.core.NameID;
import org.w3c.dom.Element;

public class PolicyManager {
    private Logger log;
    private PolicyCore core;
    private String defaultQualifier;
    
    public PolicyManager(){
        this.log = Logger.getLogger(this.getClass());
        this.core = PolicyCore.getInstance();
        PropHandler propHandler = new PropHandler("oscars.properties");
        Properties props = propHandler.getPropertyGroup("default", true);
        this.defaultQualifier = props.getProperty("domain");
        if(this.defaultQualifier == null){
            this.defaultQualifier = "";
        }else{
            this.defaultQualifier = "@"+this.defaultQualifier;
        }
    }
    
    public CheckPolicyResponse checkPolicy(CheckPolicyRequest<Requester,Element[]> request) throws RemoteException{
        CheckPolicyResponse response = new CheckPolicyResponse();
        HashMap<String,PolicyClass> policyClassMap = new HashMap<String,PolicyClass>();
        
        //Find unique policy classes for each requester
        for(Requester requester : request.getRequesters()){
            List<PolicyClass> policyClasses = this.findPolicyClass(requester);
            for(PolicyClass policyClass:policyClasses){
                if(!policyClassMap.containsKey(policyClass.getName())){
                    policyClassMap.put(policyClass.getName(), policyClass);
                    PolicyClass parent = policyClass.getParent();
                    while(parent != null && (!policyClassMap.containsKey(parent.getName()))){
                        policyClassMap.put(parent.getName(), parent);
                        parent = parent.getParent();
                    }
                }
            }
        }
        
        //Evaluate rules
        RuleEvaluator evaluator = new RuleEvaluator(this.core.getPolicyDbName());
        for(PolicyClass policyClass : policyClassMap.values()){
            String reason = evaluator.evaluate(policyClass.getRuleSet().getRule(), 
                    request.getAction(), request.getResources());
            if(reason != null && policyClass.getAllowOnMatch()){
                //deny if don't match rules
                response.setAllow(false);
                response.setReason(reason);
                return response;
            }else  if(reason == null && (!policyClass.getAllowOnMatch())){
                //deny if do match rules
                response.setAllow(false);
                response.setReason("Request does not meet policy criteria");
                return response;
            }
        }
        
        //allowed if reach this point
        response.setAllow(true);
        return response;
    }

    private List<PolicyClass> findPolicyClass(Requester requester) throws RemoteException {
        this.log.debug("findPolicyClass.start");
        String nameIdVal = requester.getSubject().getNameID().getValue();
        List<String> names = new ArrayList<String>();
        String oscarsName = null;
        PolicyClassDAO policyClassDAO = new PolicyClassDAO(this.core.getPolicyDbName());
        AaaRmiClient aaaRmiClient = this.createAaaRmiClient();
        
        /* Convert X509 Subject to username or vice versa */
        String nameFormat = requester.getSubject().getNameID().getFormat();
        String qualifier = requester.getSubject().getNameID().getNameQualifier();
        if(qualifier != null && !qualifier.startsWith("@")){
            qualifier = "@"+qualifier;
        }
        if(NameID.X509_SUBJECT.equals(nameFormat)){
            this.log.debug("X509 Name in subject");
            names.add(nameIdVal);
            names.add(this.reverseDN(nameIdVal));
            this.log.debug("added " + nameIdVal);
            try{
                oscarsName = aaaRmiClient.verifyDN(nameIdVal);
                if(oscarsName != null){
                    names.add(oscarsName+this.defaultQualifier);
                    this.log.debug("added " + oscarsName+this.defaultQualifier);
                }
            }catch(Exception e){
                this.log.debug("user not found");
            }
        }else if(Requester.OSCARS_LOGIN_SUBJECT.equals(nameFormat)){
            this.log.debug("OSCARS Name in subject");
            oscarsName = nameIdVal;
            if(qualifier == null || this.defaultQualifier.equals(qualifier)){
                //local user account
                names.add(oscarsName + this.defaultQualifier);
                this.log.debug("added " + oscarsName + this.defaultQualifier);
                try{
                    HashMap<String,Object> userQueryRequest = new HashMap<String,Object>();
                    userQueryRequest.put("objectType", ModelObject.USER);
                    userQueryRequest.put("operation", ModelOperation.FIND);
                    userQueryRequest.put("findBy", "username");
                    userQueryRequest.put("username", oscarsName);
                    HashMap<String,Object> userQueryResponse = aaaRmiClient.manageAaaObjects(userQueryRequest);
                    User user = (User)userQueryResponse.get("user");
                    if(user != null && user.getCertSubject() != null && !"".equals(user.getCertSubject().trim())){
                        names.add(user.getCertSubject());
                        names.add(this.reverseDN(user.getCertSubject()));
                        this.log.debug("added " + user.getCertSubject());
                    }
                }catch(Exception e){}
            }else{
                //remote user account
                names.add(oscarsName + qualifier);
                this.log.debug("added " + oscarsName + qualifier);
            }
        }else{
            throw new RemoteException("Invalid NameID format: " + nameFormat);
        }
        
        /* Query attributes */
        List<Attribute> attrs = null;
        if(oscarsName != null){
            HashMap<String,Object> attrQueryRequest = new HashMap<String,Object>();
            attrQueryRequest.put("objectType", ModelObject.ATTRIBUTE);
            attrQueryRequest.put("operation", ModelOperation.LIST);
            attrQueryRequest.put("listBy", "username");
            attrQueryRequest.put("username", oscarsName);
            HashMap<String,Object> attrQueryResponse = aaaRmiClient.manageAaaObjects(attrQueryRequest);
            attrs = (List<Attribute>) attrQueryResponse.get("attributes");
            this.log.debug("attrs.size=" + attrs.size());
        }
        
        /* Retrieve policy classes */
        List<PolicyClass> policyClasses = policyClassDAO.findByNameAttrs(names, attrs);
        this.log.debug("policyClasses.size="+policyClasses.size());
        
        //If no classes found use default
        if(policyClasses.isEmpty()){
            PolicyClass defaultClass = policyClassDAO.queryByParam("name", "default");
            if(defaultClass == null){
                throw new RemoteException("No policy class matches request");
            }
            policyClasses.add(defaultClass);
        }
        
        this.log.debug("findPolicyClass.end");
        return policyClasses;
    }
    
    public AaaRmiClient createAaaRmiClient() throws RemoteException{
        AaaRmiClient aaaRmiClient = new AaaRmiClient();
        try {
            aaaRmiClient.init();
        } catch (RemoteException e) {
            this.log.error("Verify that your aaa-core is running!");
            throw new RemoteException("The server was unable " +
                        "to reach an AAA service to authenticate your request");
        }
        return aaaRmiClient;
    }
    
    public String reverseDN(String dn){
        String[] dnElems = dn.split(",");
        String reverseDN = " " + dnElems[0];
        for (int i = 1; i < dnElems.length; i++) {
            reverseDN = dnElems[i] + "," + reverseDN;
        }
        reverseDN = reverseDN.substring(1);
        
        return reverseDN;
    }
}
