package net.es.oscars.policy.idc;

import java.io.File;
import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Properties;
import java.util.Set;

import javax.xml.namespace.QName;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import org.apache.axiom.om.OMAbstractFactory;
import org.apache.axis2.util.XMLUtils;
import org.apache.log4j.Logger;
import org.opensaml.saml2.core.NameID;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import net.es.oscars.ConfigFinder;
import net.es.oscars.PropHandler;
import net.es.oscars.client.Client;
import net.es.oscars.policy.PolicyCore;
import net.es.oscars.policy.RuleEvaluator;
import net.es.oscars.policy.client.Requester;
import net.es.oscars.policy.database.LTypeDAO;
import net.es.oscars.policy.database.Rule;
import net.es.oscars.policy.database.RuleDAO;
import net.es.oscars.policy.database.RuleVar;
import net.es.oscars.policy.database.RuleVarOutputParam;
import net.es.oscars.policy.ext.PolicyExtension;
import net.es.oscars.rmi.aaa.AaaRmiClient;
import net.es.oscars.wsdlTypes.ListReply;
import net.es.oscars.wsdlTypes.ListRequest;
import net.es.oscars.wsdlTypes.ResDetails;

public class IDCPolicyExt implements PolicyExtension{
    private Logger log;
    private Client client;
    private RuleDAO ruleDAO;
    private LTypeDAO lTypeDAO;
    private RuleEvaluator evaluator;
    
    private static String url = null;
    private static String repo = null;
    private static String axis2Xml = null;
    
    final static public String IDC_NS = "http://oscars.es.net/OSCARS";
    final static public String IDC_NS_PREFIX = "idc";
    final static public String STATUS_XPATH = "/idc:resDetails/idc:status";
    final static public String USER_XPATH = "/idc:resDetails/idc:login";
    final static public String BANDWIDTH_XPATH = "/idc:resDetails/idc:bandwidth";
    
    public IDCPolicyExt(){
        this.log = Logger.getLogger(this.getClass());
    }
    
    public void init() throws RemoteException{
        if(url == null){
            PropHandler propHandler = new PropHandler("oscars.properties");
            Properties props = propHandler.getPropertyGroup("idc", true);
            url = props.getProperty("url");
            if(url == null){
                this.log.debug("idc.url not defined in oscars.properties");
                url = "http://127.0.0.1:8080/axis2/services/OSCARS";
            }
        }
        if(axis2Xml == null){
           axis2Xml = ConfigFinder.getInstance().find(ConfigFinder.AXIS_TOMCAT_DIR, "axis2.xml");
           System.setProperty("axis2.xml", axis2Xml);
        }
        if(repo == null){
            repo = (new File(axis2Xml)).getParent();
        }
        
        this.ruleDAO = new RuleDAO(PolicyCore.getInstance().getPolicyDbName());
        this.lTypeDAO = new LTypeDAO(PolicyCore.getInstance().getPolicyDbName());
        this.evaluator = new RuleEvaluator(PolicyCore.getInstance().getPolicyDbName());
        this.client = new Client();
        this.client.setUp(true, url, repo, axis2Xml);
    }

    public String evaluate(Rule rule, RuleVar ruleVar, List<Requester> requesters)
            throws RemoteException {
        this.log.debug("evaluate.start");
        ListRequest request = new ListRequest();
        String reason = null;
        
        ArrayList<Rule> ruleStack = new ArrayList<Rule>();
        ruleStack.add(ruleVar.getInputRule());
        while(!ruleStack.isEmpty()){
            Rule currRule = ruleStack.remove(0);
            String lType = currRule.getLType().getName();
            if("RULE".equals(lType)){
                ruleStack.add(this.ruleDAO.queryByParam("id", currRule.getLValue()));
                ruleStack.add(this.ruleDAO.queryByParam("id", currRule.getRValue()));
                continue;
            }else if(STATUS_XPATH.equals(currRule.getLValue())){
                this.log.debug("Added status " + currRule.getRValue());
                request.addResStatus(currRule.getRValue());
            }
        }
        
        //get reservations
        ListReply reply = null;
        try {
            reply = client.listReservations(request);
        } catch (Exception e) {
            this.log.error("Error listing reservations: " + e.getMessage());
            return "Unable to get reservation list";
        }
        
        //apply expression tree
        ArrayList<Element> matchingResvs = new ArrayList<Element>();
        
        if(reply.getResDetails() != null){
            for(ResDetails resDetails : reply.getResDetails()){
                ArrayList<Element[]> elemList = new ArrayList<Element[]>();
                Element elem = null;
                try {
                    QName resDetailsQN = new QName(IDC_NS, "resDetails", IDC_NS_PREFIX);
                    elem = XMLUtils.toDOM(resDetails.getOMElement(
                            resDetailsQN, OMAbstractFactory.getOMFactory()));
                    Element[] elemArr = new Element[1];
                    elemArr[0] = elem;
                    elemList.add(elemArr);
                    
                    NodeList list = elem.getChildNodes();
                    this.log.debug("elem.getNodeName()="+elem.getNodeName());
                    this.log.debug("list.getLength()="+list.getLength());
                    for(int i = 0; i < list.getLength();i++){
                        this.log.debug(list.item(i).getNodeName());
                    }
                } catch (Exception e) {
                    this.log.error("Error converting past reservation to DOM: " + e.getMessage());
                    throw new RemoteException("Problem retrieving past reservations");
                }
                String matchString = this.evaluator.evaluate(ruleVar.getInputRule(), "", elemList, requesters);
                if(matchString == null){
                    matchingResvs.add(elem);
                }
            }
        }
        
        //Determine function
        if("sum".equals(ruleVar.getOutputType())){
            reason = this.sum(matchingResvs, ruleVar.getOutputParams(), rule, requesters);
        }
        
        this.log.debug("evaluate.end");
        
        return reason;
    }

    private String sum(ArrayList<Element> resvs, Set<RuleVarOutputParam> outputParams, Rule rule, List<Requester> requesters) throws RemoteException {
        XPath xpath = XPathFactory.newInstance().newXPath();
        xpath.setNamespaceContext(this.evaluator.getNamespaceContext());
        HashMap<String, Double> userSumMap = new HashMap<String, Double>();
        
        //get login for each user and init sum to 0
        for(Requester requester : requesters){
            this.log.info("req format="+requester.getSubject().getNameID().getFormat());
            this.log.info("req name="+requester.getSubject().getNameID().getValue());
            if(Requester.OSCARS_LOGIN_SUBJECT.equals(requester.getSubject().getNameID().getFormat())){
                userSumMap.put(requester.getSubject().getNameID().getValue(), 0.0);
            }else if(NameID.X509_SUBJECT.equals(requester.getSubject().getNameID().getFormat())){
                //get login
                AaaRmiClient aaaRmiClient = this.createAaaRmiClient();
                try{
                    String oscarsName = aaaRmiClient.verifyDN(requester.getSubject().getNameID().getValue());
                    this.log.debug("oscarsLogin="+oscarsName);
                    if(oscarsName != null){
                        userSumMap.put(oscarsName, 0.0);
                    }
                }catch(Exception e){}
            }
        }
        
        for(Element resv : resvs){
            try {
                String login = (String) xpath.evaluate(USER_XPATH, resv, XPathConstants.STRING);
                if(login != null && userSumMap.containsKey(login)){
                    Double sum = userSumMap.get(login);
                    sum += (Double) xpath.evaluate(BANDWIDTH_XPATH, resv, XPathConstants.NUMBER);
                    userSumMap.put(login, sum);
                    this.log.debug(login+".sum="+sum);
                }
            } catch (XPathExpressionException e) {
                this.log.error(e.getMessage());
                throw new RemoteException("Error retrieving bandwidth history");
            }
        }
        
        Rule modRule = new Rule();
        modRule.setLType(this.lTypeDAO.queryByParam("name", "XPATH_NUMBER"));
        modRule.setLValue(BANDWIDTH_XPATH);
        modRule.setOperator(rule.getOperator());
        modRule.setRValue(rule.getRValue());
        
        for(Double sum : userSumMap.values()){
            String reason = this.evaluator.evaluateNumber(rule, sum);
            if(reason != null){
                return reason;
            }
        }
        return null;
    }

    public void destroy() throws RemoteException {
        this.client.cleanUp();
    }
    
    public AaaRmiClient createAaaRmiClient() throws RemoteException{
        AaaRmiClient aaaRmiClient = new AaaRmiClient();
        try {
            aaaRmiClient.init();
        } catch (RemoteException e) {
            this.log.error("Verify that your aaa-core is running!");
            throw new RemoteException("The server was unable " +
                        "to reach an AAA service to authenticate your request");
        }
        return aaaRmiClient;
    }
}
